/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or
 * http://eu-egee.org/license.html
 */

/*
 * Copyright (c) 2001, 2006 EU DataGrid.
 * For license conditions see http://www.eu-datagrid.org/license.html
 *
 * Copyright (c) 2001, 2002, 2006 by
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Gerben Venekamp <venekamp@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 */

/*
 * Copyright (c) 2011, FERMI NATIONAL ACCELERATOR LABORATORY
 * All rights reserved. 
 *
 * For details of the Fermitools (BSD) license see LICENSE-Fermitools or
 *  http://fermitools.fnal.gov/about/terms.html
 */

/*!
    \page lcmaps_glexec_tracking.mod GLEXEC_TRACKING plugin

    \section posixsynopsis SYNOPSIS
    \b lcmaps_glexec_tracking.mod

    \section posixdesc DESCRIPTION

The GLEXEC_TRACKING module uses the Condor procd to setup tracking of the processes spawned by the binary using gLExec.

*/


/*!
    \file   lcmaps_glexec_tracking.c
    \brief  Plugin to execute the glexec process tracking script and return a secondary GID.
    \author Brian Bockelman for CMS.  Modified from the GUMS plugin done by Oscar Koeroo for the EGEE.  Significantly expanded by Dave Dykstra for Fermilab.
*/





/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcmaps_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>

#include <netdb.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "lcmaps/lcmaps_modules.h"
#include "lcmaps/lcmaps_arguments.h"
#include "lcmaps/lcmaps_cred_data.h"
#include "lcmaps/lcmaps_vo_data.h"

#include "lcmaps_glexec_tracking.h"

/******************************************************************************
                                Definitions
******************************************************************************/

#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1


#define CONFIG_DATA_ALLOCSIZE  512


/******************************************************************************
                          Module specific prototypes
******************************************************************************/
static int plugin_run_or_verify(int, lcmaps_argument_t *, int);


/******************************************************************************
                       Define module specific variables
******************************************************************************/
/*
 * extern char **environ;
 */

static char *execname   = NULL;
static char *config_data = NULL;
static size_t config_data_used;
static size_t config_data_alloced;
static int min_groupid = 0;
static int max_groupid = 0;

/******************************************************************************
Function:   put_config_data
Description:
    Add a keyword/value pair to config_data
Parameters:
    keyword
    value
Returns:
    LCMAPS_MOD_SUCCESS : success
    LCMAPS_MOD_FAIL    : failure because there was a malloc error
******************************************************************************/
static int put_config_data(char *keyword, char *value)
{
  size_t keylen = strlen(keyword);
  size_t vallen = strlen(value);
  size_t addlen = keylen + vallen + 2;

  if (config_data == NULL) {
    config_data_used = 0;
    config_data_alloced = 0;
  }
  
  if ((config_data_used + addlen + 1) > config_data_alloced) {
    while ((config_data_used + addlen + 1) > config_data_alloced)
      config_data_alloced += CONFIG_DATA_ALLOCSIZE;

    if (config_data == NULL) {
      config_data = (char *) malloc(config_data_alloced);
    }
    else {
      char *new_data = (char *) realloc(config_data, config_data_alloced);
      if (new_data == NULL) {
	free(config_data);
	config_data = NULL;
      }
      else
        config_data = new_data;
    }
    if (config_data == NULL)
      return LCMAPS_MOD_FAIL;
  }

  // copy in "keyword=value\n"
  strcpy(&config_data[config_data_used], keyword);
  config_data_used += keylen;
  config_data[config_data_used++] = '=';
  strcpy(&config_data[config_data_used], value);
  config_data_used += vallen;
  config_data[config_data_used++] = '\n';
  config_data[config_data_used] = '\0';

  return LCMAPS_MOD_SUCCESS;
}

/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : success
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(int argc, char **argv)
{
  char * logstr = "\tlcmaps_plugins_glexec_tracking-plugin_initialize()";
  int i;
  int got_datetime_local = 0;
  int got_dont_kill_leftovers = 0;
  char *procddir = NULL;
  char *min_gid = "65000";
  char *max_gid = "65049";
  char *pathbuf;
  int dirlen;

  for (i=0; i<argc; i++) {
    lcmaps_log_debug(2, "%s: arg %d is %s\n", logstr, i, argv[i]);

    if ((strncasecmp(argv[i], PLUGIN_GLEXEC_TRACKING_EXEC, sizeof(PLUGIN_GLEXEC_TRACKING_EXEC)) == 0) && ((i+1) < argc)) {
      if ((argv[i+1] != NULL) && (strlen(argv[i+1]) > 0)) {
        execname = strdup(argv[++i]);
        lcmaps_log_debug(2, "%s: %s has %s\n", logstr, PLUGIN_GLEXEC_TRACKING_EXEC, execname);
      }
    }
    else if ((strncasecmp(argv[i], PLUGIN_GLEXEC_TRACKING_PROCDDIR, sizeof(PLUGIN_GLEXEC_TRACKING_PROCDDIR)) == 0) && ((i+1) < argc)) {
      if ((argv[i+1] != NULL) && (strlen(argv[i+1]) > 0)) {
	++i;
        lcmaps_log_debug(2, "%s: %s has %s\n", logstr, PLUGIN_GLEXEC_TRACKING_PROCDDIR, argv[i]);

	procddir = argv[i];
      }
    }
    else if ((strncasecmp(argv[i], PLUGIN_GLEXEC_TRACKING_LOGDIR, sizeof(PLUGIN_GLEXEC_TRACKING_LOGDIR)) == 0) && ((i+1) < argc)) {
      if ((argv[i+1] != NULL) && (strlen(argv[i+1]) > 0)) {
	++i;
        lcmaps_log_debug(2, "%s: %s has %s\n", logstr, PLUGIN_GLEXEC_TRACKING_LOGDIR, argv[i]);
	if (put_config_data("log_destination", "file") == LCMAPS_MOD_FAIL)
	  goto memory_error;
	if (put_config_data("log_dir", argv[i]) == LCMAPS_MOD_FAIL)
	  goto memory_error;
      }
    }
    else if (strncasecmp(argv[i], PLUGIN_GLEXEC_TRACKING_DATETIME_LOCAL, sizeof(PLUGIN_GLEXEC_TRACKING_DATETIME_LOCAL)) == 0) {
      lcmaps_log_debug(2, "%s: got %s\n", logstr, PLUGIN_GLEXEC_TRACKING_DATETIME_LOCAL);
      got_datetime_local = 1;
    }
    else if ((strncasecmp(argv[i], PLUGIN_GLEXEC_TRACKING_MIN_GID, sizeof(PLUGIN_GLEXEC_TRACKING_MIN_GID)) == 0) && ((i+1) < argc)) {
      if ((argv[i+1] != NULL) && (strlen(argv[i+1]) > 0)) {
	++i;
        lcmaps_log_debug(2, "%s: %s has %s\n", logstr, PLUGIN_GLEXEC_TRACKING_MIN_GID, argv[i]);
	min_gid = argv[i];
      }
    }
    else if ((strncasecmp(argv[i], PLUGIN_GLEXEC_TRACKING_MAX_GID, sizeof(PLUGIN_GLEXEC_TRACKING_MAX_GID)) == 0) && ((i+1) < argc)) {
      if ((argv[i+1] != NULL) && (strlen(argv[i+1]) > 0)) {
	++i;
        lcmaps_log_debug(2, "%s: %s has %s\n", logstr, PLUGIN_GLEXEC_TRACKING_MAX_GID, argv[i]);
	max_gid = argv[i];
      }
    }
    else if (strncasecmp(argv[i], PLUGIN_GLEXEC_TRACKING_DONT_KILL_LEFTOVERS, sizeof(PLUGIN_GLEXEC_TRACKING_DONT_KILL_LEFTOVERS)) == 0) {
      lcmaps_log_debug(2, "%s: got %s\n", logstr, PLUGIN_GLEXEC_TRACKING_DONT_KILL_LEFTOVERS);
      got_dont_kill_leftovers = 1;
    }
  }

  if (execname == NULL ) {
    lcmaps_log (0, "%s Error: Required parameter %s missing\n", logstr, PLUGIN_GLEXEC_TRACKING_EXEC);
    return LCMAPS_MOD_FAIL;
  }

  if (procddir == NULL) {
    procddir = "/usr";
    dirlen = 4;
  }
  else
    dirlen = strlen(procddir);
  pathbuf = (char *) malloc(dirlen + sizeof("/sbin/condor_procd") + 1);
  if (pathbuf == NULL) goto memory_error;
  strcpy(pathbuf, procddir);
  strcpy(&pathbuf[dirlen], "/sbin/condor_procd");
  if (access(pathbuf, X_OK) != 0) {
    strcpy(&pathbuf[dirlen], "/sbin/procd");
    if (access(pathbuf, X_OK) != 0) {
      free(pathbuf);
      lcmaps_log_debug(0, "%s: Neither sbin/condor_procd nor sbin/procd found under %s.\n", logstr, procddir);
      return LCMAPS_MOD_FAIL;
    }
  }
  if (put_config_data("procd_exe", pathbuf) == LCMAPS_MOD_FAIL) {
    free(pathbuf);
    goto memory_error;
  }
  strcpy(&pathbuf[dirlen], "/sbin/gidd_alloc");
  if (access(pathbuf, X_OK) != 0) {
    free(pathbuf);
    lcmaps_log_debug(0, "%s: sbin/gidd_alloc not found under %s.\n", logstr, procddir);
    return LCMAPS_MOD_FAIL;
  }
  if (put_config_data("gidd_alloc_exe", pathbuf) == LCMAPS_MOD_FAIL) {
    free(pathbuf);
    goto memory_error;
  }
  strcpy(&pathbuf[dirlen], "/sbin/procd_ctl");
  if (access(pathbuf, X_OK) != 0) {
    strcpy(&pathbuf[dirlen], "/bin/procd_ctl");
    if (access(pathbuf, X_OK) != 0) {
      free(pathbuf);
      lcmaps_log_debug(0, "%s: procd_ctl found in neither sbin nor bin under %s.\n", logstr, procddir);
      return LCMAPS_MOD_FAIL;
    }
  }
  if (put_config_data("procd_ctl_exe", pathbuf) == LCMAPS_MOD_FAIL) {
    free(pathbuf);
    goto memory_error;
  }
  free(pathbuf);

  if (put_config_data("datetime_local", got_datetime_local ? "1" : "0") == LCMAPS_MOD_FAIL)
    goto memory_error;

  if (put_config_data("min_gid", min_gid))
    goto memory_error;

  if (put_config_data("max_gid", max_gid))
    goto memory_error;

  sscanf(min_gid, "%d", &min_groupid);
  sscanf(max_gid, "%d", &max_groupid);

  if (put_config_data("kill_leftover_daemons", got_dont_kill_leftovers ? "0" : "1") == LCMAPS_MOD_FAIL)
    goto memory_error;

  lcmaps_log_debug(1, "%s: Initialization succeeded\n", logstr);

  return LCMAPS_MOD_SUCCESS;

memory_error:
  lcmaps_log_debug(0, "%s: Memory allocation error\n", logstr);

  return LCMAPS_MOD_FAIL;
}


/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : success
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(int *argc, lcmaps_argument_t **argv)
{
  char *logstr = "\tlcmaps_plugins_glexec_tracking-plugin_introspect()";
  static lcmaps_argument_t argList[] = {
    {"pem_string",       "char *"                ,  1, NULL},
    { "user_dn"        , "char *"                ,  1, NULL},
    { "voms_data_list" ,  "lcmaps_vomsdata_t *"  ,  0, NULL},
    { "nvoms_data"     ,  "int"                  ,  0, NULL},
    {NULL        ,  NULL    , -1, NULL}
  };

  lcmaps_log_debug(2, "%s: introspecting\n", logstr);

  *argv = argList;
  *argc = lcmaps_cntArgs(argList);
  lcmaps_log_debug(1, "%s: address first argument: 0x%x\n", logstr, argList);

  lcmaps_log_debug(1, "%s: Introspect succeeded\n", logstr);

  return LCMAPS_MOD_SUCCESS;
}




/******************************************************************************
Function:   plugin_run
Description:
    Gather credentials for LCMAPS
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(int argc, lcmaps_argument_t * argv)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}


/******************************************************************************
Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) passwd/group files, etc. etc.
    The outcome should be identical to that of plugin_run().
    In this particular case "plugin_verify()" is identical to "plugin_run()"

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(int argc, lcmaps_argument_t * argv)
{
  return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv, int lcmaps_mode)
{
  FILE *fh = NULL;
  char *logstr = NULL;
  char *datetime = NULL;
  char *encoded_dn = "(void)";
  int p2c[2], c2p[2];
  time_t clock;
  struct tm *tmpTime;
  int rc = 0,rc2 = 0;
  int uid, gid, uid_count;
  int * uid_array;
  int status;
  int exit_code = -1;
  int i;
  pid_t pid;
  int ngroups;
  gid_t gidlist[10];
  char *pem_string, *dn_string, **dn_string2, *fqan_string;
  int glexec_monitor_argc;
  char ** glexec_monitor_argv;
  char * ppid_str, * uid_str;
  lcmaps_vomsdata_t * vomsdata, **vomsdata2;

  struct passwd * user_info = NULL;
  struct group  * gr_info   = NULL;


  time(&clock);
  tmpTime = gmtime(&clock);
  datetime = malloc(sizeof(char) * 20);
  snprintf(datetime, 20, "%04d-%02d-%02d %02d:%02d:%02d",
           tmpTime->tm_year + 1900, tmpTime->tm_mon + 1, tmpTime->tm_mday,
           tmpTime->tm_hour, tmpTime->tm_min, tmpTime->tm_sec);

  if (lcmaps_mode == PLUGIN_RUN)
    logstr = "\tlcmaps_plugins_glexec_tracking-plugin_run()";
  else if (lcmaps_mode == PLUGIN_VERIFY)
    logstr = "\tlcmaps_plugins_glexec_tracking-plugin_verify()";
  else {
    lcmaps_log(0, "\tlcmaps_plugins_glexec_tracking-plugin_run_or_verify(): attempt to run plugin in invalid mode: %d\n", lcmaps_mode);

    goto glexec_tracking_failure;
  }

  if ((pem_string = *(char **) lcmaps_getArgValue("pem_string", "char *", argc, argv))) {
    lcmaps_log_debug(1, "%s: found PEM string\n", logstr);
  } else {
    lcmaps_log(0, "%s: could not get value of PEM string.\n", logstr);
    goto glexec_tracking_failure;
  }

  lcmaps_log_debug(2, "%s: Acquiring information from LCMAPS framework\n", logstr);
  dn_string2 = (char **)lcmaps_getArgValue("user_dn", "char *",argc, argv);
  if ((dn_string2 == NULL) || ((dn_string = *dn_string2) == NULL)) {
    lcmaps_log_debug(1, "%s: value of user_dn is empty. No user DN found by the framework in the proxy chain.\n", logstr);
    goto glexec_tracking_failure;
  } else {
    lcmaps_log_debug(5, "%s: user_dn = %s\n", logstr, dn_string);
  }
   
  lcmaps_log_debug(5,"%s: Grabbing VOMS data\n", logstr);
  vomsdata = NULL;
  vomsdata2 = (lcmaps_vomsdata_t **)lcmaps_getArgValue("voms_data_list", "lcmaps_vomsdata_t *", argc, argv);
  if ((vomsdata2 == NULL) || ((vomsdata = *vomsdata2) == NULL)) {
    lcmaps_log_debug(5, "%s: No VOMS data; will not pass VOMS info to glexec_monitor.\n", logstr);
  } else {
    lcmaps_log_debug (5, "%s: vomsdata->nvoms = %d\n", logstr, vomsdata->nvoms);
    lcmaps_log_debug (5, "%s: vomsdata->voms = %x\n", logstr, vomsdata->voms);
    lcmaps_log_debug (5, "%s: vomsdata->voms[0] = %x\n", logstr, vomsdata->voms[0]);
  }

  uid_count = 0;
  uid_array = (int *)getCredentialData(UID, &uid_count);
  if (uid_count != 1) {
    lcmaps_log(0, "%s: No UID set yet; must map to a UID before running the glexec_tracking module.\n", logstr);
    goto glexec_tracking_failure;
  }
  uid = uid_array[0];

  pipe(p2c);
  pipe(c2p);

  pid = fork();
  switch(pid) {
    //  Child Process
    case 0:
      close(p2c[1]);
      close(c2p[0]);
      dup2(p2c[0], 0);
      dup2(c2p[1], 1);
      close(p2c[0]);
      close(c2p[1]);

      //  Setting the real and effective uid/gid to root.
      if (setreuid(0, 0) != 0) {
        lcmaps_log_debug(0, "%s: Setting the real/effective uid to root failed.\n", logstr);
      }

      if (setregid(0, 0) != 0) {
        lcmaps_log_debug(0, "%s: Setting the real/effective gid to root failed.\n", logstr);
      }

      //  Build the argument strings
      int glexec_monitor_argc=7;
      if (vomsdata != NULL) {
        glexec_monitor_argc += 2 + vomsdata->voms[0].nfqan;
      }
 
      glexec_monitor_argv = (char **)calloc(glexec_monitor_argc + 1, sizeof(char *));   //  Allocate the terminating NULL pointer also!
      ppid_str = (char *)calloc(MAX_INT_STR, sizeof(char));
      uid_str = (char *)calloc(MAX_INT_STR, sizeof(char));
      if ((glexec_monitor_argv == NULL) || (ppid_str == NULL) || (uid_str == NULL)) {
          lcmaps_log(0, "%s: Memory allocation error.\n", logstr);
          goto glexec_tracking_failure;
      }
      if (snprintf(ppid_str, MAX_INT_STR, "%d", getppid()) < 0) {
          lcmaps_log(0, "%s: Unable to write out the PPID.\n", logstr);
          goto glexec_tracking_failure;
      }
      
      if (snprintf(uid_str, MAX_INT_STR, "%d", uid) < 0) {
          lcmaps_log(0, "%s: Unable to write out the UID.\n", logstr);
          goto glexec_tracking_failure;
      }
      glexec_monitor_argv[0] = "glexec_monitor";
      glexec_monitor_argv[1] = "-c";
      glexec_monitor_argv[2] = config_data;
      glexec_monitor_argv[3] = ppid_str; // PPID
      glexec_monitor_argv[4] = ""; // Job ID
      glexec_monitor_argv[5] = uid_str; // UID
      glexec_monitor_argv[6] = dn_string;
      if (vomsdata != NULL) {
        glexec_monitor_argv[7] = vomsdata->voms[0].voname;
        glexec_monitor_argv[8] = vomsdata->voms[0].voms_issuer_dn;
        for (i=0; i<vomsdata->voms[0].nfqan; i++) {
          glexec_monitor_argv[9+i] = vomsdata->voms[0].fqan_unix[i].fqan;
        }
      }

      lcmaps_log(5, "%s: execing %s\n", logstr, execname);
      for (i=0; i < glexec_monitor_argc; i++) {
        lcmaps_log(5, "%s: param %d: %s\n", logstr, i, glexec_monitor_argv[i]);
      }
      
      execve(execname, glexec_monitor_argv, NULL);

      //  Above function should not return. If it does, there is an error.
      lcmaps_log (0, "%s Error: Unable to execute glexec_monitor found at: %s\n", logstr, execname);
      goto glexec_tracking_failure;

      break;

    //  Error
    case -1:
      goto glexec_tracking_failure;
      break;

    //  Parent process
    default:
      close(p2c[0]);
      close(c2p[1]);
  }

  //  Dump the string to stdout first and close it when finished.
  write(p2c[1], pem_string, strlen(pem_string));
  close(p2c[1]);

  //  After dumping the string to stdout, we need to expect an answer
  //  back on stdin.
  fh = fdopen(c2p[0], "r");
  rc2 = fscanf(fh, "%d", &gid);

  fclose(fh);

  wait(&status);

  exit_code = WEXITSTATUS(status);

  if (exit_code != 0)
  {
     lcmaps_log (0, "Error: failure in execute. Exit code %d. Executed \"%s\".\n", exit_code, execname);

     /* failure in child */
     goto glexec_tracking_failure;
  }

  if (rc2!=1) {
     lcmaps_log (0, "%s Error: invalid return %d. Expected a tracking GID.\n", logstr,rc);
     goto glexec_tracking_failure;
  }

  lcmaps_log (5, "%s: Adding SEC_GID :  %d\n", logstr, gid);
  rc+=addCredentialData(SEC_GID,&gid);
  ngroups=getgroups(sizeof(gidlist)/sizeof(gid_t),&gidlist[0]);
  for(i=0;i<ngroups;i++) {
    if ((gidlist[i] >= min_groupid) && (gidlist[i] <= max_groupid)) {
      lcmaps_log (5, "%s: Inheriting SEC_GID :  %d\n", logstr, gidlist[i]);
      rc+=addCredentialData(SEC_GID,&gidlist[i]);
    } else {
      lcmaps_log (5, "%s: Not inheriting secondary gid %d, outside of range\n", logstr, gidlist[i]);
    }
  }

  if (rc==0) {
     lcmaps_log_time(0, "%s: glexec_tracking plugin succeeded\n", logstr);
  } else {
     lcmaps_log (0, "%s Error: At least one of the addCredentialData calls failed.\n", logstr);
     goto glexec_tracking_failure;
  }

  if (datetime) free (datetime);
  return LCMAPS_MOD_SUCCESS;


glexec_tracking_failure:
  if (datetime) free (datetime);
  lcmaps_log_time(0, "%s: glexec_tracking plugin failed\n", logstr);

  return LCMAPS_MOD_FAIL;
}


/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : success
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate()
{
  char * logstr = "\tlcmaps_plugins_glexec_tracking-plugin_terminate()";
  lcmaps_log_debug(1,"%s: terminating\n", logstr);

  // Free stuff
  if (execname) {
    free(execname);
    execname = NULL;
  }
  if (config_data) {
    free(config_data);
    config_data = NULL;
  }

  return LCMAPS_MOD_SUCCESS;
}
